# -*- coding: iso-8859-1 -*-
#
# Copyright (c) 2006 Martin Lesser <ml@bettercom.de>, All Rights Reserved
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2 or later, the text of which can be found at
# http://www.gnu.org/copyleft/gpl.html. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# $Id$

import cairo, locale, math

__version__ = '0.1.1'
__version_info__ = (0, 1, 1)

def _hexcolor(s=''):
    """Return the rgb-values (float) for a html-like colorstring
    (i.e. #FF0000 = red)"""
    r, g, b = 0, 0, 0
    try:
        if s[0] == '#' and len(s) == 7:
            r, g, b = (float('0x%s' % s[1:3]),
                       float('0x%s' % s[3:5]),
                       float('0x%s' % s[5:7]))
    except:
        pass
    return r/255, g/255, b/255

def _num2str(a, dec=0, t_sep=True, perc=False, slocal='de_DE'):
    """Format an int or float as 'pretty' string"""
    if isinstance(a, str):
        return a
    try:
        _local=locale.getlocale(locale.LC_ALL)
    except:
        _local=(None,None)
    if not _local[0] or _local[0] <> slocal:
        for _l in ('', '.utf8', '.iso8859-1'):
            try:
                locale.setlocale(locale.LC_ALL, '%s%s' % (slocal, _l))
                break
            except:
                pass
    try:
        fstr = '%%.%df' % dec
        n = locale.format(fstr, a, t_sep)
        if perc:
            n = '%s %%' % n
    except:
        n = str(a)
    if _local[0] and _local[0] <> slocal:
        # Reset to old locale
        locale.setlocale(locale.LC_ALL, _local)
    elif not _local[0]:
        locale.setlocale(locale.LC_ALL, 'C')
    return n

def _rnd_y(_v=0):
    """Calc a 'pretty' y_max"""
    _x = max(1, int(10 ** int(math.ceil(math.log10(_v)) - 2)))
    return int(math.ceil(float(_v) / _x) * _x), _x

class Chart:

    def __init__(self, **kv):
        """New instance of Graph with some default attributes if not
        set explicit"""
        self.width     = kv.get('width', 200)
        self.height    = kv.get('height', 200)
        # margin is used at several places for spacing and padding:
        # - distance between elements
        # - linespace etc.
        self.margin    = kv.get('margin', 5)
        self.linewidth = float(int(kv.get('linewidth', 1)))

        # Type of chart (bar or curve)
        self._type = kv.get('type', 'bar').lower()

        # Intersection of y and x-axis
        # (if -1 then the lowest data-value will be taken)
        self.x0 = kv.get('x0', 0)

        # area contains x_min, y_min, x_max and y_max of the
        # grid where we may paint the graph
        self.area      = [self.margin, self.margin,
                          self.width - self.margin, self.height - self.margin]

        # The labels for the axis (x, y, [y2]):
        self.labels = {}
        # The data
        self.data = []

        # A list with columns to be stacked:
        self._stack = []
        # A list with columns for which a second y-axis at the
        # right side should be built (as transparent curve)
        # XXX: Should we stack them?
        self._y2 = []

        # Instantiate img and context of cairo-class
        self._setupcairo(kv.get('out',''))

        # Defaultfont
        self.defaultfont = kv.get('font', {'name' : 'Sans', 'size' : 10,
                                           'bold' : False, 'italic' : False})
        self.font = {}
        self.set_font(**self.defaultfont)

        # The locale for formatting numbers
        self.locale = kv.get('locale', 'en_US')

        # Some colors:
        # The current color
        self.color = None
        # a dict with all colors used
        self.colors = {}
        # list with data-colors, will be filled in set_legend
        # TODO: Define some default-colors
        self.dcolors = []
        # Set background-color (white if none is given):
        self.set_color(kv.get('colors',{}).get('bg', '#FFFFFF'), alias='bg')
        # Fill Graph with background-color:
        self.rectangle(0, 0, self.width, self.height)
        # Set grid-color (light grey if none is given
        if 'grid' in kv.get('colors', {}) and not kv.get('colors',{}).get('grid'):
            # grid-lines explicitly disabled
            pass
        else:
            self.set_color(kv.get('colors',{}).get('grid','#aaaaaa'), alias='grid')
        # Set foreground-color (black if not given):
        self.set_color(kv.get('colors',{}).get('fg', '#000000'), alias='fg')
        # Draw border
        self.ctx.set_line_width(self.linewidth)
        if 'border' in kv.keys() and not kv.get('border'):
            # No border wanted
            pass
        else:
            self.rectangle(0, 0, self.width, self.height, outline=True)

        # Set True if you Don't want to draw labels and data-values if val==0:
        self.suppress_0s = False

        # Do we need these both really?
        self.x, self.y = 0, 0

        # Maximum, stepwidth and correction-factor of y-axis:
        # (first value in each list for the left axis,
        #  second value for right y-axis
        self.y_max, self.y_step, self.y_corr = [1, 1], [1, 1], [1, 1]

    def _setupcairo(self, out=''):
        """Instantiate surface and context depending on the filetype"""
        self.file = out
        if isinstance(out, str) and out.endswith('.pdf'):
            if not cairo.HAS_PDF_SURFACE:
                raise SystemExit ('cairo was not compiled with PDF support')
            self._out_method  = self._show_page
            self.img   = cairo.PDFSurface(self.file,
                                          self.width, self.height)
            self.img.set_dpi(72.0, 72.0)
        elif isinstance(out, str) and out.endswith('.ps'):
            if not cairo.HAS_PS_SURFACE:
                raise SystemExit ('cairo was not compiled with PS support')
            self._out_method  = self._show_page
            self.img   = cairo.PSSurface(self.file,
                                         self.width, self.height)
            self.img.set_dpi(72.0, 72.0)
        elif isinstance(out, str) and out.endswith('.svg'):
            # pycairo 1.0.2 has no member SVG...
            #if not cairo.HAS_SVG_SURFACE:
            raise SystemExit ('cairo was not compiled with SVG support')
            #self._out_method  = self._show_page
            #self.img   = cairo.SVGSurface(self.file,
            #                             self.width, self.height)
        elif isinstance(out, (str, file)):
            self._out_method = self._write_to_png
            self.img   = cairo.ImageSurface(cairo.FORMAT_ARGB32,
                                            self.width, self.height)
        else:
            raise SystemExit ('CairoChart needs a filename or -object to write to')
        self.ctx       = cairo.Context(self.img)
        self.matrix    = self.ctx.get_matrix()

    def set_color(self, _n='', **kv):
        """Set color"""
        if kv.get('values'):
            (r, g, b, a) = self.color = kv.get('values')
            self.ctx.set_source_rgba(r, g, b, a)
            return
        if _n and self.colors.get(_n):
            r, g, b, a = self.colors.get(_n)
        elif _n:
            r, g, b = _hexcolor(_n)
            a = kv.get('alpha', 1)
            if kv.get('alias'):
                self.colors[kv.get('alias')] = (r, g, b, a)
            self.colors[_n] = (r, g, b, a)
        else:
            r, g, b, a = 0, 0, 0, 1
        if kv.get('alpha', 1) <> a:
            a = kv.get('alpha')
        if (r, g, b, a) == self.color:
            return
        self.set_color(values=(r, g, b, a))

    def set_data(self, data=[], **kv):
        """Set data and calculate the 'drawable' area where
        the data-values will be drawn.
        
        Also calc the heigth and width of x- and y-
        lines, grid and labels which will be drawn outside of
        this 'drawable'"""
        self._stack = kv.get('stack', [])
        self._y2 = kv.get('y2', [])
        self.set_font(**self.defaultfont)
        # The text-height of each x-label
        # (which will become the width due to rotating):
        lw = h = self.textsize(str(data[0][0])).get('max_h')
        # The x-labels will be turned by 90 so the height of each label
        # (before rotating) is the default width of each rectangle
        # and the width of the widest label (before rotating) becomes
        # the height of the x-labels
        # With colwidth we may define another distance between the
        # x-labels:
        if kv.get('colwidth', 0):
            h = kv.get('colwidth')
            if h < 0:
                h = lw
        # _margin is the distance between the columns:
        _margin = kv.get('margin', self.margin)
        # The maximum number of values (colums) to show
        # (before moving left and right y-axis):
        _max_x_vals = int((self.area[2] - self.area[0]) / (h + _margin))
        # Set x-labels and move y-pos of x-axis up so the x-labels will fit below
        self.area[3] -= (self.margin + self._set_x_labels(data[-_max_x_vals:]))

        # Set y-labels and move x-pos of y-axis to the right
        self.area[0] += (self.margin + self._set_y_labels())
        # If we have data2 we need a second y-axis
        # at the right site of the drawable and shrink the
        # drawable at the right side
        if self._y2:
            self.area[2] -= (self.margin + self._set_y_labels(right=True))
        # Finetune the width of one total column with 1 or more sub-columns:
        width = float(self.area[2] - self.area[0] -
                      _margin * (_max_x_vals - 1)) / _max_x_vals
        if kv.get('align', 'l')[0].lower() == 'r':
            # Align drawing-area at the right side
            _n = (self.area[2] - self.area[0] -
                  len(data) * width - (len(data)-1) * _margin)
            _n = max(0.0, _n)
        else:
            _n = 0.0
        # Now we can calculate x, y and width of the x-labels:
        for _x in self.labels.get('x'):
            _x[2] = self.area[0] + _n
            _x[3] = self.area[3] + self.margin + _x[1]
            _x[4] = width
            _n += (width + _margin)

    def set_font(self, **font):
        """Set the font"""
        self.font['italic'] = font.get('italic', False)
        self.font['bold']   = font.get('bold', False)
        self.font['name']   = font.get('name',
                                       self.font.get('name', 'Sans'))
        self.ctx.select_font_face(self.font.get('name'),
                                  self.font.get('italic'),
                                  self.font.get('bold'))
        if font.get('size') and font.get('size') <> self.font.get('size'):
            self.ctx.set_font_size(font.get('size'))
            self.font['size'] = font.get('size')

    def set_legend(self, legend=[], **kv):
        """Draw/Write legend to bottom of chart"""
        self.set_font(**kv.get('font', {'size' : 12}))
        _w = self.margin
        _barwidth = kv.get('barwidth', 30)
        rows = 1
        row_h = self.textsize(str(legend[0][0])).get('max_h')
        for _e in legend:
            # First calc the number of lines to print
            row_w = self.textsize(str(_e[0])).get('width') + self.margin * 2 + _barwidth
            if row_w > (self.width - 2 * self.margin):
                rows += 1
                _w = self.margin
            elif (_w + row_w) > self.width or _e[0][0] == '\n':
                rows += 1
                _w =  self.margin + row_w
            else:
                _w += row_w
        _y_legend = int(self.area[3] - rows * (row_h + self.margin))
        self.area[3] -= (self.margin + rows * (row_h + self.margin))
        self.set_color('fg')
        self.line([(0, _y_legend), (self.width, _y_legend)])
        _y_legend +=  self.margin
        _x = self.margin
        for _e in legend:
            row_w = self.textsize(str(_e[0])).get('width')
            if (_x + _barwidth + row_w + self.margin) > self.width or _e[0][0] == '\n':
                _x = self.margin
                _y_legend += (row_h + self.margin)
            if _e[0][0] == '\n':
                _t = _e[0][1:]
            else:
                _t = _e[0]
            self.set_color(_e[1])
            self.dcolors.append(_e[1])
            self.rectangle(_x, _y_legend, _barwidth, row_h)
            _x += self.margin + _barwidth
            self.set_color('fg')
            self.text(_t, _x, _y_legend)
            _x += (row_w + self.margin)

    def set_title(self, text='', **kv):
        """Create the title-bar with text 'text' and a
        bottom-line below the centered text.
        If no color is given FG-color will be used.
        If no font is given the default-font bold 18 will
        be used."""
        self.set_font(**kv.get('font', {'bold':True, 'size':18}))
        self.set_color(kv.get('color', 'fg'))
        _h = self.textsize(str(text)).get('max_h')
        _y = self.margin
        for _t in str(text).split('\n'):
            self.text(_t, self.width / 2, 0 + _y, align='c')
            _y += _h
        _y += self.margin
        self.line([(0, _y), (self.width, _y)])
        self.area[1] = _y + self.margin
        self.set_font(**self.defaultfont)
        if kv.get('copyright'):
            self.text(kv.get('copyright'), self.width - self.margin,
                      (self.height - self.margin -
                       self.textsize(kv.get('copyright')).get('max_h')),
                      align='r')

    def draw_bars(self):
        """Draw the data as bar-chart"""
        _col = 0
        for _d in self.labels.get('x'):
            _data = self.data[_col]
            _x = _d[2]
            if not self._stack:
                _c = 0
                # The width of one sub-column:
                _s_width = _d[4] / (len(_data) - len(self._y2))
                for _v in _data:
                    if (_c + 1) in self._y2:
                        continue
                    try:
                        self.set_color(self.dcolors[_c])
                    except:
                        self.set_color('fg')
                    _h = _v * self.y_corr[0]
                    _y = self.area[3] - _h
                    self.rectangle(_x, _y, _s_width, _h)
                    _x += _s_width
                    _c += 1
            else:
                _s_width = _d[4] / len(self._stack)
                for _s in self._stack:
                    _y = self.area[3]
                    for _c in _s:
                        try:
                            self.set_color(self.dcolors[_c-1])
                        except:
                            self.set_color('fg')
                        _h = _data[_c-1] * self.y_corr[0]
                        _y -= _h
                        self.rectangle(_x, _y, _s_width, _h)
                    _x += _s_width
            _col += 1

    def draw_data(self, **kv):
        """Draw the data as rectangle or whatever else..."""
        if self._type == 'bar':
            self.draw_bars()
            if self._y2:
                self.draw_lines(**kv)
        elif self._type == 'line':
            self.draw_lines(**kv)

    def draw_dots(self, points=[], **kv):
        """Draw dots with radius of self.linewidth at the given points"""
        _radius = kv.get('radius', self.linewidth)
        for _x0, _y0 in points:
            self.ctx.arc(_x0, _y0, _radius, 0, math.radians(360))
            self.ctx.fill()

    def draw_grid(self, **kv):
        """Draw both (optional three) axis,
        the labels of each axis and the grid"""
        label_h = self.textsize('X').get('max_h')
        _n = 0
        _y_max_len = max([_x[1] for _x in self.labels.get('y')])
        for _l in self.labels.get('y'):
            # X-Pos for each y-label:
            _l[1] = self.margin + _y_max_len - _l[1]
            # Y-Pos:
            _l[2] = self.area[1] + _n * self.y_step[0] - label_h / 2
            _p = self.text(_l[0], self.margin + _y_max_len,
                           _l[2], align='r')
            _n += 1
            if _n < len(self.labels.get('y')):
                if not self.colors.get('grid'):
                    continue
                self.set_color('grid')
                _p = self.line([(self.area[0], _l[2] + label_h / 2),
                                (self.area[2], _l[2] + label_h / 2)],
                               dash=[3,2])
                self.set_color('fg')
        if self.labels.get('y2'):
            _y_max_len = max([_x[1] for _x in self.labels.get('y2')])
            _n = 0
            # Change color here:
            self.set_color(self.dcolors[self._y2[0]-1])
            for _l in self.labels.get('y2'):
                # X-Pos for each y-label:
                _l[1] = self.width - self.margin - _y_max_len
                # Y-Pos:
                _l[2] = self.area[1] + _n * self.y_step[1] - label_h / 2
                _p = self.text(_l[0], _l[1],
                               _l[2], align='l')
                _n += 1
            self.set_color('fg')
        self.line([(self.area[0]-self.linewidth, self.area[1] - label_h / 2),
                   (self.area[0]-self.linewidth, self.area[3]),
                   (self.area[2]+self.linewidth, self.area[3]),
                   (self.area[2]+self.linewidth, self.area[1] - label_h / 2)])
        # Adjust grid with linewidths of x/y axis:
        #self.area[0] += 1
        self.area[1] += self.linewidth
        #self.area[3] -= 0.5
        self.y_corr[0] = (self.area[3] - self.area[1]) / float(self.y_max[0])
        self.y_corr[1] = (self.area[3] - self.area[1]) / float(self.y_max[1])
        self.suppress_0s = kv.get('suppress_0s', False)
        for _l in self.labels.get('x'):
            if self.suppress_0s and _l[6] == 0:
                # Don't draw x-label and rectangle if only zero-vals
                continue
            # Draw text (x-label), adjust _x:
            self.set_color('fg')
            _x = _l[2] + (_l[4] - label_h) / 2
            self.text(_l[0], _x, _l[3], rotate=270)

    def draw_lines(self, **kv):
        """Draw data as line-chart"""
        for _l in range(0, len(self.data[0])):
            _col = 0
            if (_l + 1) not in self._y2 and self._type <> 'line':
                continue
            _lp = []
            if (_l+1) in self._y2:
                _axis=1
            else:
                _axis=0
            for _d in self.labels.get('x'):
                _x = _d[2] + _d[4] / 2
                _y = self.area[3] - self.data[_col][_l] * self.y_corr[_axis] - self.linewidth
                if not _col and not self.suppress_0s:
                    # Begin line at y0-axis
                    _lp.append((_d[2], _y))
                if not self.suppress_0s or self.data[_col][_l]:
                    _lp.append((_x, _y))
                _col += 1
            if not self.suppress_0s:
                _lp.append((_x + _d[4] / 2, _y))
            try:
                self.set_color(self.dcolors[_l])
            except:
                self.set_color('fg')
            self.line(_lp, dots=True)
            if kv.get('dots'):
                self.draw_dots(_lp)

    def _set_x_labels(self, data):
        """Set up a list with the labels of x-axis and
        calculate the maximum-value for each entry.
        Return the width of the widest x-label in pixels"""
        # Each element in the x-labels-list is a list with 8 values:
        # [text, width of text, x-pos, y-pos, maximum width, maximum height,
        # maximum y-value, maximum y2-value]
        self.labels['x'] = [[str(_x[0]),
                             self.textsize(str(_x[0])).get('width'),
                             0, 0, 0, 0]
                            for _x in data]
        self.data = [_x[1:] for _x in data]
        _c, _l = 0, len(self.data[0])
        for _v in self.data:
            _m0, _m1 = 0, 0
            if not self._stack and not self._y2:
                _m0 = max(_v)
            elif not self._stack:
                for _n in range(0, _l):
                    if (_n + 1) in self._y2:
                        _m1 = max(_m1, _v[_n])
                    else:
                        _m0 = max(_m0, _v[_n])
            else:
                _s = 1
                _m2 = [0] * len(self._stack)
                for _e in _v:
                    if _s in self._y2:
                        _m1 += _e
                    else:
                        _z = 0
                        for _ze in self._stack:
                            if _s in _ze:
                                _m2[_z] += _e
                            _z += 1
                        _m0 = max(_m2)
                    _s += 1
            self.labels['x'][_c].append(_m0)
            self.labels['x'][_c].append(_m1)
            _c += 1
                
        # Return text-width in pixels of the widest x-label:
        return max([_x[1] for _x in self.labels.get('x')])

    def _set_y_labels(self, right=False):
        """Calculate the greatest value on y-axis and setup
        the stepwidth between y-labels.
        Return the width of the widest y-label in pixels"""
        if right:
            _axis, _label = 1, 'y2'
        else:
            _axis, _label = 0, 'y'
        _y_max = max([_x[6 + _axis] for _x in self.labels.get('x')])
        self.y_max[_axis], self.y_step[_axis] = _rnd_y(int(_y_max))
        # Height of y-label:
        self.set_font(**self.defaultfont)
        label_h = self.textsize(_num2str(self.y_max[_axis],
                                         slocal=self.locale)).get('max_h')
        y_step = self.y_step[_axis]
        while (self.y_max[_axis] / y_step * label_h) > (self.area[3] - self.area[1]):
            # Enlarge step-size if y-labels would overlap
            y_step += self.y_step[_axis]
        self.y_step[_axis] = y_step
        # Recalc max. y-label so all steps have the same value:
        self.y_max[_axis] = int(math.ceil(float(self.y_max[_axis]) / y_step) *
                                y_step)
        _l_texts = [_num2str(_y, slocal=self.locale)
                    for _y in range(self.y_max[_axis], -1,
                                    - self.y_step[_axis])]
        self.labels[_label] = [[_y, self.textsize(_y).get('width'), 0]
                               for _y in _l_texts]
        if not _axis:
            self.area[1] += (label_h / 2 + self.linewidth)
        # The distance in pixels between the y-labels:
        self.y_step[_axis] = ((self.area[3] - self.area[1]) /
                              (len(self.labels.get(_label)) - 1))
        # Return width of the widest y-label:
        try:
            return max([_x[1] for _x in self.labels.get(_label)])
        except:
            return 0

    def line(self, points=[], **kv):
        """Draw a line starting at the first (x,y)-tuple in points
        through all other x/y-tuples in points. Try to fix them before
        so no 'smearing' will appear.  Possible parameters are:

        width: linewidth If no width is given self.linewidth will be used

        dash: A dash pattern is specified by dashes, a list of
        positive values. Each value provides the user-space length of
        alternate 'on' and 'off' portions of the stroke.

        linecap:  butt (default), round, square
        linejoin: miter (default), round, bevel
        """
        if len(points) < 2:
            return
        if not kv.get('width'):
            _width = self.linewidth
        else:
            _width = float(int(kv.get('width')))
        self.ctx.set_line_width(_width)
        # Fix 'smearing'-problem:
        _width = float(int(_width) % 2) / 2
        if kv.get('dash'):
            self.ctx.set_dash(kv.get('dash'), 1)
        else:
            self.ctx.set_dash([], 0)
        if kv.get('linecap','b')[0] == 'r':
            self.ctx.set_line_cap(cairo.LINE_CAP_ROUND)
        elif kv.get('linecap','b')[0] == 's':
            self.ctx.set_line_cap(cairo.LINE_CAP_SQUARE)
        else:
            self.ctx.set_line_cap(cairo.LINE_CAP_BUTT)
        if kv.get('linejoin','m')[0] == 'r':
            self.ctx.set_line_join(cairo.LINE_JOIN_ROUND)
        elif kv.get('linejoin','m')[0] == 'b':
            self.ctx.set_line_join(cairo.LINE_JOIN_BEVEL)
        else:
            self.ctx.set_line_join(cairo.LINE_JOIN_MITER)
        x0, y0 = points.pop(0)
        start = True
        for x1,y1 in points:
            _x0, _x1, _y0, _y1 = (float(int(x0)), float(int(x1)),
                                  float(int(y0)), float(int(y1)))
            if x0 == x1:
                # vertical line, fix _x0 and _x1 by linewidth / 2
                _x0 += _width
                _x1 += _width
                _y1 += _width
            elif y0 == y1:
                # horizontal line, fix yx0 and yx1 by linewidth / 2
                _y0 += _width
                _y1 += _width
                _x1 += _width
            if start:
                self.ctx.move_to(_x0, _y0)
                start = False
            self.ctx.line_to(_x1, _y1)
            x0, y0 = x1, y1
        self.ctx.stroke()

    def rectangle(self, _x=0, _y=0, _w=0, _h=0, **kv):
        """Draw a rectangle at _x, _y with width _w and height _h
        If outline is True stroke() instead of fill() will be used
        and _x, _y etc. will be fixed so no 'smearing' appears"""
        if kv.get('outline'):
            _x += self.linewidth / 2
            _y += self.linewidth / 2
            _w -= self.linewidth
            _h -= self.linewidth
        self.ctx.rectangle(_x, _y, _w, _h)
        if kv.get('outline'):
            if kv.get('dash'):
                self.ctx.set_dash(kv.get('dash'), 1)
            else:
                self.ctx.set_dash([], 0)
            self.ctx.stroke()
        else:
            self.ctx.fill()

    def text(self, _text='', _x=0, _y=0, **kv):
        """Print text '_text' at _x/_y
        ctx.text_path() normally draws text above the baseline of the
        text and ctx.rectangle below _y.
        So to get this in 'sync' we also draw text below x/y so
        _x and _y define the left upper corner of the bounding box
        and not x/y of the leftmost part of the baseline.
        
        Currently the following parameters may be given:
        rotate: the text will be rotated by x degrees
        align:  the text will be aligned (rigth, left, center)
        valign: vertical align (top [default], middle, bottom, baseline [cairo-default])
        border: Print a surrounding border"""
        if kv.get('font'):
            self.set_font(**kv.get('font'))
        if kv.get('color'):
            self.set_color(**kv.get('color'))
        _size = self.textsize(_text)
        _r = math.radians(kv.get('rotate', 0))
        if kv.get('align', 'l').lower()[0] == 'r':
            _a = _size.get('width')
        elif kv.get('align', 'l').lower()[0] == 'c':
            _a = _size.get('width') / 2
        else:
            _a = 0
        if kv.get('valign', 't').lower()[0] == 'l':
            # keep cairo-behavior drawing from baseline
            _v = 0
        elif kv.get('valign', 't').lower()[0] == 'm':
            # adjust _x,_y so the middle of the the text is
            # center for rotation 
            _v = _size.get('max_h') / 2 - _size.get('max_d')
        elif kv.get('valign', 't').lower()[0] == 'b':
            # move text to bottom-line of bounding-box
            _v = -_size.get('max_d')
        else:
            # Our default is setting the top-left corner of bounding-box
            # at _x, _y
            _v = _size.get('max_a')
        self.ctx.move_to(_x, _y)
        # Move _x, _y so the top-left corner of bounding-box is at _x, _y
        self.ctx.rel_move_to(-math.sin(_r) * _v, math.cos(_r) * _v)
        self.ctx.rotate(_r)
        self.ctx.rel_move_to(-_a,0)
        _bx, _by = self.ctx.get_current_point()
        _by -= _size.get('max_a')
        self.ctx.text_path(_text)
        self.ctx.fill()
        if kv.get('border'):
            self.rectangle(_bx, _by, _size.get('width'), _size.get('max_h'), outline=True)
        if _r:
            self.ctx.set_matrix(self.matrix)

    def textsize(self, _text='', **kv):
        """Calculate the width and height of a text
        Read http://www.cairographics.org/manual/cairo-Scaled-Fonts.html#cairo-text-extents-t
        and http://www.cairographics.org/manual/cairo-Scaled-Fonts.html#cairo-font-extents-t
        as reference"""
        if kv.get('font'):
            self.set_font(**kv.get('font'))
        _l = self.ctx.text_extents(_text)
        _f = self.ctx.font_extents()
        return {'width'  : _l[4], # Width of bounding box
                'height' : _l[3], # Height of bounding box
                'max_h'  : _f[2], # Total height of the font
                'max_a'  : _f[0], # ascent, max height of glyphs above baseline
                'max_d'  : _f[1]  # descent of the font below the baseline
                }
        
    def output(self):
        """Write Image to self.file, _file may be a filename or a file-like object"""
        if not self.file:
            return False
        try:
            self._out_method()
        except:
            raise #SystemExit ('output() failed')

    def _write_to_png(self):
        self.img.write_to_png(self.file)

    def _show_page(self):
        self.ctx.show_page()
        self.img.finish()
